local EventHandler = FlexBar2.OO.Mixin
{
	-- Custom event interface functions, act as advanced (and tweaked) versions of the AceEvent ones
	"RegisterEvent", -- Custom RegisterEvent function, Registers the event with AceEvent if it ain't already. It is table-based, read its code for details ;)
	"UnregisterEvent",
	"UnregisterAllEvents",

	"PreClick_RegisterCastingEvent", -- Register casting events so the button is notified from casting or channeling (this way, only clicked buttons are notified and memory is spared so not all butons do the check)
	"PostClick_UncheckButton", -- Uncheck the default checking of the button after it has been clicked
	"UNIT_INVENTORY_CHANGED_UpdateGearChecks", -- Takes take of checking the buttons in the equipped color
	"ACTIONBAR_SLOT_CHANGED_UpdateTextures", -- Update button icons when an actionbar slot changed
	"PLAYER_ENTERING_WORLD_LoadButtonTextures", -- Update the button icons after all information is available to the UI
	"ACTIONBAR_UPDATE_COOLDOWN_SetSpinner", -- Set the cooldown spinner
	"SpellCast_Start", -- Triggered whenever a spellcast starts (whenever a button is clicked)
	"SpellCast_End", -- Triggered whenever a spellcast ends, be it on the SPELLCAST_FAILED or SPELLCAST_STOP event
	"Channeling_Start", -- Triggered when the player starts to channel a spell
	"Channeling_Stop", -- Triggered when the player stops channeling a spell
	"Buff_Update", -- Triggered when the player gains or loses a buff
	"Debuff_Gained", -- Triggered when the target gains a debuff
	"Debuff_Lost", -- Triggered when the target loses a debuff
	"Debuff_Update", -- Triggered when the player's target changes and that debuffs have to be updated
}


function EventHandler:Activate()
	self.Frame.Button:SetScript("PreClick", self.PreClick_RegisterCastingEvent);
	self.Frame.Button:SetScript("PostClick", self.PostClick_UncheckButton);
	self:RegisterEvent("ACTIONBAR_SLOT_CHANGED", "Buff_Update");
	self:RegisterEvent("ACTIONBAR_SLOT_CHANGED", "ACTIONBAR_SLOT_CHANGED_UpdateTextures");
	self:RegisterEvent("PLAYER_ENTERING_WORLD", "PLAYER_ENTERING_WORLD_LoadButtonTextures");
	self:RegisterEvent("SpecialEvents_UnitDebuffGained", "Debuff_Gained");
	self:RegisterEvent("SpecialEvents_UnitDebuffLost", "Debuff_Lost");
	self:RegisterEvent("SpecialEvents_AuraTargetChanged", "Debuff_Update");
	self:RegisterEvent("ACTIONBAR_UPDATE_COOLDOWN", "ACTIONBAR_UPDATE_COOLDOWN_SetSpinner");
	if(type(self.Settings) == "table" and type(self.Settings.Attributes) == "table" and self.Settings.Attributes.type == "action") then -- Trigger some functions to have buffs & worn gearn checked correctly
		self:Buff_Update(nil, self.Settings.Attributes.action);
		self:UNIT_INVENTORY_CHANGED_UpdateGearChecks(nil, "player");
	end
end


function EventHandler:RegisterEvent(Event, MethodName)
	FlexBar2:Debug("Registering " , Event, " with " , MethodName);
	if(type(self[MethodName]) ~= "function") then
		return false
	end
	if(not FlexBar2:IsEventRegistered(Event)) then
		FlexBar2[Event] = function(...)
			for n ,u in pairs(FlexBar2.EventTable[Event]) do
				for _,v in pairs(u) do
					n[v](n, ...);
				end
			end
		end
		FlexBar2:RegisterEvent(Event);
	end
	if(type(FlexBar2.EventTable[Event]) ~= "table") then
		FlexBar2.EventTable[Event] = {};
	end
	if(type(FlexBar2.EventTable[Event][self]) ~= "table") then
		FlexBar2.EventTable[Event][self] = {};
	end
	table.insert(FlexBar2.EventTable[Event][self], MethodName);
end

function EventHandler:UnregisterEvent(Event, MethodName)
	if(MethodName) then
		return FlexBar2:RemoveByValue(FlexBar2.EventTable[Event][self], MethodName);
	else
		for n,v in pairs(FlexBar2.EventTable[Event][self]) do
			table.remove(FlexBar2.EventTable[Event][self], n);
		end
	end
end


function EventHandler:PLAYER_ENTERING_WORLD_LoadButtonTextures(Addon)
	FlexBar2:Debug("PLAYER_ENTERING_WORLD_LoadButtonTextures");
	self:LoadCurrentTexture();
end

function EventHandler:PreClick_RegisterCastingEvent()
	FlexBar2:Debug("PreClick_RegisterCastingEvent");
	local self = this:GetClass();
	self:RegisterEvent("UNIT_SPELLCAST_START", "SpellCast_Start");
	self:RegisterEvent("UNIT_SPELLCAST_CHANNEL_START","Channeling_Start");
end

function EventHandler:PostClick_UncheckButton()
	if(this:GetClass().Check[1] ~= nil) then
		this:SetChecked(true);
	else
		this:SetChecked(false);
	end
end

function EventHandler:UNIT_INVENTORY_CHANGED_UpdateGearChecks(Addon, Unit)
	if(Unit == "player" and type(self.Settings) == "table" and type(self.Settings.Attributes) == "table" and self.Settings.Attributes.type == "action"
	and HasAction(self.Settings.Attributes.action)
	and IsEquippedAction(self.Settings.Attributes.action) == 1) then
		self:SetChecked(true, "Equipped");
	else
		self:SetChecked(false, "Equipped");
	end
end

function EventHandler:ACTIONBAR_SLOT_CHANGED_UpdateTextures(Addon, Slot)
	-- Update actionbar slot texture if this button's slot corresponds with the changed slot(s)
	if(type(self.Settings) == "table" and type(self.Settings.Attributes) == "table" and self.Settings.Attributes.type == "action" and self.Settings.Attributes.action == Slot) then
		FlexBar2:Debug("ACTIONBAR_SLOT_CHANGED_UpdateTextures");
		FlexBar2:TriggerEvent("ACTIONBAR_SLOT_CHANGED_UpdateTextures");
		self:LoadCurrentTexture();
	end
end

function EventHandler:ACTIONBAR_UPDATE_COOLDOWN_SetSpinner(Addon, MouseButton)
	if(self.Settings and self.Settings.Attributes and self.Settings.Attributes.type == "action") then
		local start, duration, enable = GetActionCooldown(self.Settings.Attributes.action);
		if(self.LastCooldown.start ~= start or self.LastCooldown.duration ~= duration) then
			self.LastCooldown.start = start;
			self.LastCooldown.duration = duration;
			if(FlexBar2.db.profile.Ticker == "global") then
				if(duration <= 1.5) then
					CooldownFrame_SetTimer(self.Frame.Button.Cooldown, start, duration, enable);
				end
			elseif(FlexBar2.db.profile.Ticker == "noglobal") then
				if(duration > 1.5) then
					CooldownFrame_SetTimer(self.Frame.Button.Cooldown, start, duration, enable);
				end
			elseif(FlexBar2.db.profile.Ticker == true) then
				CooldownFrame_SetTimer(self.Frame.Button.Cooldown, start, duration, enable);
			end
		end
	end
end

function EventHandler:SpellCast_Start(Addon, Unit)
	if(Unit == "player" and self.Settings and self.Settings.Attributes and self.Settings.Attributes.type == "action" and IsCurrentAction(self.Settings.Attributes.action)) then
		FlexBar2:Debug("SpellCast_Start"); -- Trigger the SpellCast_Start event and check the button
		FlexBar2:TriggerEvent("SpellCast_Start");
		self:SetChecked(true, "Casting");

		self:UnregisterEvent("UNIT_SPELLCAST_START", "SpellCast_Start");
		self:UnregisterEvent("UNIT_SPELLCAST_CHANNEL_START","Channeling_Start");
		self:RegisterEvent("UNIT_SPELLCAST_STOP", "SpellCast_End");
	end
end

function EventHandler:SpellCast_End(Addon, Unit)
	if(Unit == "player" and self.Settings and self.Settings.Attributes and self.Settings.Attributes.type == "action") then
		FlexBar2:Debug("SpellCast_End"); -- We will take care of unchecking the button
		FlexBar2:TriggerEvent("SpellCast_End");
		self:SetChecked(false, "Casting");

		self:UnregisterEvent("UNIT_SPELLCAST_STOP", "SpellCast_End");
	end
end

function EventHandler:Channeling_Start(Addon, Unit)
	if(Unit == "player" and self.Settings and self.Settings.Attributes and self.Settings.Attributes.type == "action" and IsCurrentAction(self.Settings.Attributes.action)) then
		FlexBar2:Debug("Channeling_Start"); -- Check the button for channeling
		FlexBar2:TriggerEvent("Channeling_Start");
		self:SetChecked(true, "Channeling");

		self:UnregisterEvent("UNIT_SPELLCAST_START", "SpellCast_Start");
		self:UnregisterEvent("UNIT_SPELLCAST_CHANNEL_START","Channeling_Start");
		self:RegisterEvent("UNIT_SPELLCAST_CHANNEL_STOP", "Channeling_Stop");
	end
end

function EventHandler:Channeling_Stop(Addon, Unit)
	if(Unit == "player" and self.Settings and self.Settings.Attributes and self.Settings.Attributes.type == "action") then
		FlexBar2:Debug("Channeling_Stop"); -- We will take care of unchecking the button
		FlexBar2:TriggerEvent("Channeling_Stop");
		self:SetChecked(false, "Channeling");

		self:UnregisterEvent("UNIT_SPELLCAST_CHANNEL_STOP", "Channeling_Stop");
	end
end

function EventHandler:Buff_Update(Addon, Slot) -- We need to delay this slightly because buffs don't update immediately after the event triggers
	FlexBar2:ScheduleEvent(function()
		if(self.Settings and self.Settings.Attributes and self.Settings.Attributes.type == "action" and self.Settings.Attributes.action == Slot and HasAction(Slot)) then
			FlexBar2:Debug("Buff_Update", FlexBar2.Aura:UnitHasBuff("player", nil, nil, GetActionTexture(Slot)));
			FlexBar2:TriggerEvent("Buff_Update");
			if(FlexBar2.Aura:UnitHasBuff("player", nil, nil, GetActionTexture(Slot)) ~= nil) then
				self:SetChecked(true, "Buffed");
			else
				self:SetChecked(false, "Buffed");
			end
		end
	end, 0.2);
end

function EventHandler:Debuff_Gained(Addon, Unit, Name, Count, Type, Texture, Rank, Index)
	if(Unit == "target" and self.Settings and self.Settings.Attributes and self.Settings.Attributes.type == "action" and GetActionTexture(self.Settings.Attributes.action) == Texture) then
		self:SetChecked(true, "Debuffed");
		-- TODO: Work out a ButtonText containing the ammount of applications a debuff has
	end
end

function EventHandler:Debuff_Lost(Addon, Unit, Name, Count, Applications, Type, Texture, Rank)
	if(Unit == "target" and self.Settings and self.Settings.Attributes and self.Settings.Attributes.type == "action" and GetActionTexture(self.Settings.Attributes.action) == Texture) then
		self:SetChecked(false, "Debuffed");
	end
end

function EventHandler:Debuff_Update(Addon)
	if(self.Settings and self.Settings.Attributes and self.Settings.Attributes.type == "action" and HasAction(self.Settings.Attributes.action) and FlexBar2.Aura:UnitHasDebuff("target", nil, nil, GetActionTexture(self.Settings.Attributes.action))) then
		self:SetChecked(true, "Debuffed");
	else
		self:SetChecked(false, "Debuffed");
	end
end

FlexBar2:RegisterClass(EventHandler);
